library light_server;
import 'dart:io';
import 'dart:async';
import 'dart:mirrors';
import 'dart:math' as Math;

part 'environment.dart';
part 'utils.dart';

typedef void Called(Request req);

Env env = new Development();
String ROOT = 'ROOT';
String INDEX = 'INDEX';

class Request {
  HttpRequest req;
  Map<String, String> post;
  
  HttpResponse get resp => req.response;
  Uri get uri => req.uri;
  String get method => req.method;
  Map<String, String> get get => uri.queryParameters;
  HttpSession get session => req.session;
  
  Request(this.req, this.post);
}

class LightServer {
  Map<String, Called> called;
  
  LightServer.fromMap(this.called);
  
  LightServer.fromLibrary(Symbol l) {
    called = new Map();
    currentMirrorSystem().findLibrary(l).declarations.forEach((clSym, clMirror){
      if(clMirror is ClassMirror && clMirror.superclass.qualifiedName == #light_server.Route) {
        // We have a class
        List<String> path = new List();
        String clName = MirrorSystem.getName(clSym);
        if(clName != ROOT) path.add(clName);
        clMirror.declarations.forEach((mtSym, mtMirror){
          if(mtMirror is MethodMirror && mtMirror.isStatic) {
            // We have a method
            List<String> currentPath = new List.from(path);
            String mtName = MirrorSystem.getName(mtSym);
            if(mtName != INDEX) currentPath.add(mtName);
            // Here we define the closure handling every request.
            called[currentPath.join("/")] = (Request req){
              // Now we call the method
              List<Object> params = new List.from(req.uri.pathSegments);
              params.removeRange(0, currentPath.length);
              params.insertAll(0, [req]);
              if(params.length <= mtMirror.parameters.where((e)=>!e.isNamed).length && ParamsChecker.check(mtMirror.parameters, params)) {
                Map<String, dynamic> named = new Map();
                clMirror.invoke(mtSym, params, NamedFiller.check(mtMirror.parameters, req.post));
              } else
                req.resp.write("Not found");
            };
          }
        });
      }
    });
  }
  
  void listen(Stream<HttpRequest> stream) {
    stream.listen((HttpRequest req){
      Stopwatch w = new Stopwatch();
      w.start();
      Called c;
      List<String> page = new List.from(req.uri.pathSegments);
      while(c == null) {
        c = called[page.join("/")];
        if(page.length > 0)
          page.removeLast();
        else
          break;
      }
      if(c == null)
        req.response.write("Not found and no index");
      StringBuffer s = new StringBuffer();
      req.listen((List<int> datas){
        s.write(new String.fromCharCodes(datas));
      }, onDone: (){
        c(new Request(req, Uri.splitQueryString(s.toString())));
        env.info(w.elapsedMicroseconds);
        req.response.close();
      });
    });
  }
}